#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_stdint.h"
#include "mx_image.h"
#include "mx_bit.h"
#include "mx_area_detector.h"
#include "d_pccd_170170.h"

int
main( int argc, char *argv[] )
{
	static const char fname[] = "descramble_pccd_170170";

	FILE *file;
	MX_IMAGE_FRAME *image_frame;
	uint16_t ***pccd_170170_sector_array;
	long row_framesize, column_framesize;
	long num_sector_rows, num_sector_columns;
	long i_framesize, j_framesize;
	char *raw_image_data;
	long raw_image_length, bytes_read;
	char *raw_filename, *smv_filename;
	int saved_errno;
	mx_status_type mx_status;

	if ( argc < 5 ) {
		fprintf( stderr,
		"Usage: %s #columns #rows raw_filename smv_filename\n",
			argv[0] );
		exit(1);
	}

	column_framesize = atol( argv[1] );
	row_framesize    = atol( argv[2] );
	raw_filename     = argv[3];
	smv_filename     = argv[4];

	raw_image_length = 2 * column_framesize * row_framesize;

	raw_image_data = malloc( raw_image_length );

	if ( raw_image_data == NULL ) {
		mx_status = mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate a %ld byte "
			"raw image data array.", raw_image_length );

		exit( mx_status.code );
	}

	file = fopen( raw_filename, "r" );

	if ( file == NULL ) {
		saved_errno = errno;

		mx_status = mx_error( MXE_FILE_IO_ERROR, fname,
			"Cannot open raw image file '%s'.  "
			"Errno = %d, error message = '%s'.",
			raw_filename, saved_errno, strerror(saved_errno) );

		exit( mx_status.code );
	}

	bytes_read = fread( raw_image_data, 1, raw_image_length, file );

	if ( bytes_read < raw_image_length ) {
		mx_status = mx_error( MXE_FILE_IO_ERROR, fname,
		"The number of bytes (%ld) read from raw filename '%s' "
		"was less than the expected number of bytes (%ld).",
			bytes_read, raw_filename, raw_image_length );

		exit( mx_status.code );
	}

	/* Allocate an MX_IMAGE_FRAME to hold the descrambled data. */

	image_frame = malloc( sizeof(MX_IMAGE_FRAME) );

	if ( image_frame == (MX_IMAGE_FRAME *) NULL ) {
		mx_status = mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate "
			"an MX_IMAGE_FRAME structure." );

		exit( mx_status.code );
	}

	mx_status = mx_image_alloc( &image_frame,
				row_framesize,
				column_framesize,
				MXT_IMAGE_FORMAT_GREY16,
				mx_native_byteorder(),
				2,
				MXT_IMAGE_HEADER_LENGTH,
				raw_image_length );

	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	/* We overlay the frame data buffer in 'image_frame' with
	 * 16 sector arrays in a 4 x 4 grid.  Each of the 16 CCD
	 * taps has its own separate sector in the final descrambled
	 * image.  So you end up with something that looks like this:
	 *
	 * +-----------+-----------+-----------+-----------+
	 * |           |           |           |           |
	 * | Sector  0 | Sector  1 | Sector  2 | Sector  3 |
	 * |           |           |           |           |
	 * +-----------+-----------+-----------+-----------+
	 * |           |           |           |           |
	 * | Sector  4 | Sector  5 | Sector  6 | Sector  7 |
	 * |           |           |           |           |
	 * +-----------+-----------+-----------+-----------+
	 * |           |           |           |           |
	 * | Sector  8 | Sector  9 | Sector 10 | Sector 11 |
	 * |           |           |           |           |
	 * +-----------+-----------+-----------+-----------+
	 * |           |           |           |           |
	 * | Sector 12 | Sector 13 | Sector 14 | Sector 15 |
	 * |           |           |           |           |
	 * +-----------+-----------+-----------+-----------+
	 *
	 * Each of the 16 sectors has its own independent
	 * 2-dimensional array as an overlay onto the 
	 * 1-dimensional image data buffer.
	 */

	num_sector_rows = 4;
	num_sector_columns = 4;

	i_framesize = column_framesize / num_sector_columns;
	j_framesize = row_framesize / num_sector_rows;

	mx_status = mxd_pccd_170170_alloc_sector_array(
					&pccd_170170_sector_array,
					j_framesize, i_framesize,
					num_sector_rows, num_sector_columns,
					image_frame->image_data );

	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	/* Now we descramble the raw data into the image frame. */

	mx_status = mxd_pccd_170170_descramble_raw_data(
					(uint16_t *) raw_image_data,
					pccd_170170_sector_array,
					i_framesize, j_framesize );
	
	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	mx_status = mx_image_write_smv_file( image_frame, smv_filename );
	
	if ( mx_status.code != MXE_SUCCESS )
		exit( mx_status.code );

	exit(0);
}

